# This is a dummy module to hold resources that can be shared across processes
# It will be reloaded in each run of pygosolnp

par_lower_limit = None
par_upper_limit = None
eq_values = None
ineq_lower_bounds = None
ineq_upper_bounds = None
pysolnp_property_keys = None
pysolnp_property_values = None
pygosolnp_property_keys = None
pygosolnp_property_values = None
parameter_guesses = None

obj_func = None
eq_func = None
ineq_func = None

pysolnp_delta = None
pysolnp_rho = None
pysolnp_max_major_iter = None
pysolnp_max_minor_iter = None
pysolnp_tolerance = None
pysolnp_debug = None
evaluation_type = None
number_of_parameters = None
eval_results = None
restart_results = None
restart_convergence = None
